//-----------------------------------------------------------------------------
// Copyright (c) 2009 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------
//
// This application demonstrates how to render a 2D textured rectangle using
// OpenGL 3.1 or a forward compatible OpenGL 3.0 rendering context.
//
//-----------------------------------------------------------------------------

#include <windows.h>
#include <gdiplus.h>
#include <stdexcept>
#include "application.h"

class Demo : public GL3Application
{
public:
    Demo();
    virtual ~Demo();

protected:
    virtual void init(int argc, char *argv[]);
    virtual void render();
    virtual void shutdown();
    virtual void update(float elapsedTimeSec);

private:
    void initQuad(GLfloat width, GLfloat height);
    void initShaderProgram();
    void ortho2D(GLfloat left, GLfloat right, GLfloat bottom, GLfloat top);
    void renderQuad() const;
    void translate(GLfloat x, GLfloat y, GLfloat z);

    struct Quad
    {
        struct Vertex
        {
            GLfloat pos[2];
            GLfloat texcoord[2];
        };

        const Texture2D *pTexture;
        
        GLuint shaderProgram;

        GLuint vertexArrayObject;
        GLuint vertexBufferObject;
        GLuint indexBufferObject;

        GLushort indices[4];
        Vertex vertices[4];
    };
    
    GLfloat m_projectionMatrix[16];
    GLfloat m_viewMatrix[16];
    GLfloat m_modelMatrix[16];
    Quad m_quad;
};

//-----------------------------------------------------------------------------
// Demo class implementation.
//-----------------------------------------------------------------------------

Demo::Demo() : GL3Application("OpenGL 3 - Hello, World!")
{
    memset(&m_quad, 0, sizeof(m_quad));
}

Demo::~Demo()
{
}

void Demo::init(int argc, char *argv[])
{
    // Load the decal texture for the quad.

    m_quad.pTexture = loadTexture2D("Content\\Textures\\opengl3_logo.png", GL_LINEAR, GL_LINEAR, GL_REPEAT, GL_REPEAT);
    
    if (!m_quad.pTexture)
        throw std::runtime_error("Unable to load: opengl3_logo.png.");

    // Load the shader program used to render the quad.
    // The shader program for the quad needs to be loaded before setting up
    // the vertex array and buffer objects because user-defined generic
    // vertex shader variables are being used.

    initShaderProgram();   

    // Initialize the geometry for the quad.
    // Here we also define the mappings for the generic vertex shader variables
    // that the quad's shader program uses.

    initQuad(GLfloat(m_quad.pTexture->width), GLfloat(m_quad.pTexture->height));
    
    // Load the identity matrix into the model, view, and projection matrices.

    memset(m_projectionMatrix, 0, sizeof(m_projectionMatrix));
    m_projectionMatrix[0] = 1.0f;
    m_projectionMatrix[5] = 1.0f;
    m_projectionMatrix[10] = 1.0f;
    m_projectionMatrix[15] = 1.0f;

    memset(m_viewMatrix, 0, sizeof(m_viewMatrix));
    m_viewMatrix[0] = 1.0f;
    m_viewMatrix[5] = 1.0f;
    m_viewMatrix[10] = 1.0f;
    m_viewMatrix[15] = 1.0f;

    memset(m_modelMatrix, 0, sizeof(m_modelMatrix));
    m_modelMatrix[0] = 1.0f;
    m_modelMatrix[5] = 1.0f;
    m_modelMatrix[10] = 1.0f;
    m_modelMatrix[15] = 1.0f;
}

void Demo::initQuad(GLfloat width, GLfloat height)
{
    // The quad is centered about the world origin.

    m_quad.vertices[0].pos[0] = -width * 0.5f;
    m_quad.vertices[0].pos[1] = height * 0.5f;
    m_quad.vertices[0].texcoord[0] = 0.0f;
    m_quad.vertices[0].texcoord[1] = 1.0f;

    m_quad.vertices[1].pos[0] = -width * 0.5f;
    m_quad.vertices[1].pos[1] = -height * 0.5f;
    m_quad.vertices[1].texcoord[0] = 0.0f;
    m_quad.vertices[1].texcoord[1] = 0.0f;

    m_quad.vertices[2].pos[0] = width * 0.5f;
    m_quad.vertices[2].pos[1] = height * 0.5f;
    m_quad.vertices[2].texcoord[0] = 1.0f;
    m_quad.vertices[2].texcoord[1] = 1.0f;

    m_quad.vertices[3].pos[0] = width * 0.5f;
    m_quad.vertices[3].pos[1] = -height * 0.5f;
    m_quad.vertices[3].texcoord[0] = 1.0f;
    m_quad.vertices[3].texcoord[1] = 0.0f;

    // The quad will be rendered as a triangle strip.
    
    m_quad.indices[0] = 0;
    m_quad.indices[1] = 1;
    m_quad.indices[2] = 2;
    m_quad.indices[3] = 3;

    // Create the vertex array object for the quad.
    //
    // The buffer objects that are to be used by the vertex stage of the GL
    // are collected together to form a vertex array object. All state related
    // to the definition of data used by the vertex processor is encapsulated
    // in a vertex array object.
    //
    // Vertex array objects maintain state. Any required vertex and index
    // buffer objects are created and bound to the vertex array object. Then
    // any vertex shader generic shader input variables are mapped to the
    // vertex data stored in the vertex buffer objects bound to the vertex
    // array object. All of this only needs to be done once when using vertex
    // array objects.
    //
    // When we want to draw the quad we simple bind the vertex array object and
    // then call glDrawElements() to do the drawing. There is no need to define
    // the quad's vertex array format again since that information has already
    // been specified and saved to the vertex array object.
    //
    // See the renderQuad() method to see how vertex array objects are used in
    // practice.

    glGenVertexArrays(1, &m_quad.vertexArrayObject);
    glBindVertexArray(m_quad.vertexArrayObject);

    // Create the Index Buffer Object for the quad.

    glGenBuffers(1, &m_quad.indexBufferObject);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_quad.indexBufferObject);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(m_quad.indices), m_quad.indices, GL_STATIC_DRAW);

    // Create the Vertex Buffer Object for the quad.
    
    glGenBuffers(1, &m_quad.vertexBufferObject);
    glBindBuffer(GL_ARRAY_BUFFER, m_quad.vertexBufferObject);
    glBufferData(GL_ARRAY_BUFFER, sizeof(m_quad.vertices), m_quad.vertices, GL_STATIC_DRAW);

    // GLSL 1.30 (OpenGL 3.0) deprecates varying and attribute shader
    // variables. These are replaced with used-defined generic vertex and
    // fragment shader input and output variables.
    //
    // Map the generic vertex attributes used by the quad's shader program to
    // the quad's vertex data stored in the vertex buffer object.

    glVertexAttribPointer(glGetAttribLocation(m_quad.shaderProgram, "Position"), 2, GL_FLOAT, GL_FALSE, sizeof(Quad::Vertex), BUFFER_OFFSET(0));
    glEnableVertexAttribArray(glGetAttribLocation(m_quad.shaderProgram, "Position"));

    glVertexAttribPointer(glGetAttribLocation(m_quad.shaderProgram, "TexCoord0"), 2, GL_FLOAT, GL_FALSE, sizeof(Quad::Vertex), BUFFER_OFFSET(sizeof(GLfloat) * 2));
    glEnableVertexAttribArray(glGetAttribLocation(m_quad.shaderProgram, "TexCoord0"));
}

void Demo::initShaderProgram()
{
    if (!(m_quad.shaderProgram = loadShaderProgram("Content\\Shaders\\quad.vert", "Content\\Shaders\\quad.frag")))
    {
        if (programInfoLog().length() > 0)
            throw std::runtime_error(programInfoLog());

        if (shaderInfoLog().length() > 0)
            throw std::runtime_error(shaderInfoLog());
    }

    // When binding the quad's texture map to the shader we pass the sampler
    // uniform variable the texture unit number that the texture is bound to.
    // A common mistake is to set the sampler variable with the quad texture's
    // actual texture object id.

    glUniform1i(glGetUniformLocation(m_quad.shaderProgram, "ColorMapSampler"), 0);
}

void Demo::ortho2D(GLfloat left, GLfloat right, GLfloat bottom, GLfloat top)
{
    // Build a 2D orthographic projection matrix.
    // This is the same matrix used by gluOrtho2D() and glOrtho().

    const GLfloat nearVal = -1.0f;
    const GLfloat farVal = 1.0f;

    GLfloat tx = -(right + left) / (right - left);
    GLfloat ty = -(top + bottom) / (top - bottom);
    GLfloat tz = -(farVal + nearVal) / (farVal - nearVal);

    m_projectionMatrix[0] = (2.0f / (right - left));
    m_projectionMatrix[1] = 0.0f;
    m_projectionMatrix[2] = 0.0f;
    m_projectionMatrix[3] = 0.0f;

    m_projectionMatrix[4] = 0.0f;
    m_projectionMatrix[5] = (2.0f / (top - bottom));
    m_projectionMatrix[6] = 0.0f;
    m_projectionMatrix[7] = 0.0f;

    m_projectionMatrix[8] = 0.0f;
    m_projectionMatrix[9] = 0.0f;
    m_projectionMatrix[10] = (-2.0f / (farVal - nearVal));
    m_projectionMatrix[11] = 0.0f;

    m_projectionMatrix[12] = tx;
    m_projectionMatrix[13] = ty;
    m_projectionMatrix[14] = tz;
    m_projectionMatrix[15] = 1.0f;
}

void Demo::render()
{
    glViewport(0, 0, viewportWidth(), viewportHeight());
    glClearColor(1.0f, 1.0f, 1.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    renderQuad();
}

void Demo::renderQuad() const
{
    glUseProgram(m_quad.shaderProgram);

    // GLSL 1.30 (OpenGL 3.0) deprecates most of the built-in uniform state
    // variables. In GLSL 1.40 (OpenGL 3.1) they have been removed. We now
    // directly supply the projection, view, and model matrices to our shader.
    // OpenGL 3.0 also deprecates fixed-function vertex processing. This means
    // that we also need to calculate the projection, view, and model matrices
    // by ourself.
    glUniformMatrix4fv(glGetUniformLocation(m_quad.shaderProgram, "ProjectionMatrix"), 1, GL_FALSE, m_projectionMatrix);
    glUniformMatrix4fv(glGetUniformLocation(m_quad.shaderProgram, "ViewMatrix"), 1, GL_FALSE, m_viewMatrix);
    glUniformMatrix4fv(glGetUniformLocation(m_quad.shaderProgram, "ModelMatrix"), 1, GL_FALSE, m_modelMatrix);
        
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, m_quad.pTexture->texture);

    glBindVertexArray(m_quad.vertexArrayObject);
    glDrawElements(GL_TRIANGLE_STRIP, sizeof(m_quad.indices) / sizeof(m_quad.indices[0]), GL_UNSIGNED_SHORT, BUFFER_OFFSET(0));
    glBindVertexArray(0);

    glBindTexture(GL_TEXTURE_2D, 0);
    glUseProgram(0);
}

void Demo::shutdown()
{
    if (m_quad.vertexArrayObject)
    {
        glDeleteVertexArrays(1, &m_quad.vertexArrayObject);
        m_quad.vertexArrayObject = 0;
    }

    if (m_quad.vertexBufferObject)
    {
        glDeleteBuffers(1, &m_quad.vertexBufferObject);
        m_quad.vertexBufferObject = 0;
    }

    if (m_quad.indexBufferObject)
    {
        glDeleteBuffers(1, &m_quad.indexBufferObject);
        m_quad.indexBufferObject = 0;
    }

    glUseProgram(0);

    if (m_quad.shaderProgram)
    {
        glDeleteProgram(m_quad.shaderProgram);
        m_quad.shaderProgram = 0;
    }
}

void Demo::translate(GLfloat x, GLfloat y, GLfloat z)
{
    // Build a translation matrix.
    // This is the same matrix used by glTranslatef().

    m_modelMatrix[0] = 1.0f;
    m_modelMatrix[1] = 0.0f;
    m_modelMatrix[2] = 0.0f;
    m_modelMatrix[3] = 0.0f;

    m_modelMatrix[4] = 0.0f;
    m_modelMatrix[5] = 1.0f;
    m_modelMatrix[6] = 0.0f;
    m_modelMatrix[7] = 0.0f;

    m_modelMatrix[8] = 0.0f;
    m_modelMatrix[9] = 0.0f;
    m_modelMatrix[10] = 1.0f;
    m_modelMatrix[11] = 0.0f;

    m_modelMatrix[12] = x;
    m_modelMatrix[13] = y;
    m_modelMatrix[14] = z;
    m_modelMatrix[15] = 1.0f;
}

void Demo::update(float elapsedTimeSec)
{
    if (keyDown(Keyboard::KEY_LALT) || keyDown(Keyboard::KEY_RALT))
    {
        if (keyPressed(Keyboard::KEY_ENTER))
            toggleFullScreen();
    }

    if (keyPressed(Keyboard::KEY_ESCAPE))
        quit();

    // Update the projection and model matrices.
    // Render the quad in 2D and ensure it's centered in the viewport.

    GLfloat width = GLfloat(viewportWidth());
    GLfloat height = GLfloat(viewportHeight());

    ortho2D(0.0f, width, 0.0f, height);
    translate(width * 0.5f, height * 0.5f, 0.0f);
}

//-----------------------------------------------------------------------------
// WinMain - application entry point.
//-----------------------------------------------------------------------------

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
    Demo demo;

    demo.preferAnisotropicTextureFiltering();
    demo.preferAntiAliasing();

    return demo.run();
}